#!/usr/bin/env python

# Copyright (c) 2009 J. Lewis Muir <jlmuir@anl.gov>
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
# SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
# OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

#
# For usage, see printUsage function or run with -h option.
#

import getopt
import asynchat
import asyncore
import math
import os
import re
import signal
import socket
import sys
import time

DEFAULT_PORT = 50434

def main(args):
  (port, isDebugOutputEnabled) = parseCommandLineOptions(args)
  ConnectionDispatcher(port, isDebugOutputEnabled)
  asyncore.loop()

def getProgramName(args=None):
  if args == None: args = sys.argv
  if len(args) == 0 or args[0] == "-c": return "PROGRAM_NAME"
  return os.path.basename(args[0])

def initializeSignalHandlers():
  signalToNameMap = {
    signal.SIGHUP:"SIGHUP",
    signal.SIGINT:"SIGINT",
    signal.SIGPIPE:"SIGPIPE",
    signal.SIGALRM:"SIGALRM",
    signal.SIGTERM:"SIGTERM",
    signal.SIGUSR1:"SIGUSR1",
    signal.SIGUSR2:"SIGUSR2"
  }

  def signalHandler(signum, frame):
    if signum in [signal.SIGINT, signal.SIGTERM]:
      print
      print "%s caught. Exiting." % signalToNameMap[signum]
      sys.exit(0)
    else:
      print >> sys.stderr
      print >> sys.stderr, "%s caught. Exiting." % signalToNameMap[signum]
      sys.exit(1)

  for each in signalToNameMap.keys():
    signal.signal(each, signalHandler)

def parseCommandLineOptions(args):
  (options, extra) = getopt.getopt(args[1:], "dp:h", ["debug", "port=",
    "help"])

  if len(extra) > 0:
    print >> sys.stderr, "Error: unexpected command line option \"%s\"" % \
      extra[0]
    printUsage(sys.stderr)
    sys.exit(1)

  port = DEFAULT_PORT
  isDebugOutputEnabled = False

  for eachOptName, eachOptValue in options:
    if eachOptName in ("-d", "--debug"):
      isDebugOutputEnabled = True
    elif eachOptName in ("-p", "--port"):
      port = int(eachOptValue)
    elif eachOptName in ("-h", "--help"):
      printUsage(sys.stdout)
      sys.exit(0)

  return (port, isDebugOutputEnabled)

def printUsage(outStream):
  print >> outStream, """\
Usage: %s [OPTION]...
Options:
  -d,--debug        Print debug messages to stderr
  -p,--port=NUMBER  Listen on the specified port NUMBER for incoming
                    connections (default: %d)
  -h,--help         Print usage message and exit
""" % (getProgramName(), DEFAULT_PORT)

class ConnectionDispatcher(asyncore.dispatcher):
  def __init__(self, port, debugOutputEnabled=False):
    asyncore.dispatcher.__init__ (self)
    self.port = port
    self.debugOutputEnabled = debugOutputEnabled
    self.devices = []
    for i in xrange(16): self.devices.append(IcPlusQuad())
    self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
    self.set_reuse_addr()
    self.bind(("", port))
    self.listen(5)

  def handle_accept(self):
    ConnectionSession(self.accept(), self.devices, self.debugOutputEnabled)

class ConnectionSession(asynchat.async_chat):
  SetIcRe = re.compile(r":set(?P<ad>\d+):ic(?P<ch>\d+)\s+(?P<val>\d+)$", re.I)

  GetXyRe = re.compile(r":read(?P<ad>\d+):pos_xy\?$", re.I)
  GetCurrAllRe = re.compile(r":read(?P<ad>\d+):currall\?$", re.I)

  SetAvgCurrRe = re.compile(r":conf(?P<ad>\d+):avgcurr\s+(?P<val>\d+)$", re.I)
  SetSingleRe = re.compile(r":conf(?P<ad>\d+):single$", re.I)
  SetRangeRe = re.compile(r":conf(?P<ad>\d+):curr:rang\s+(?P<val>\d+)$", re.I)
  GetRangeRe = re.compile(r":conf(?P<ad>\d+):curr:rang\?$", re.I)

  SetGxRe = re.compile(r":conf(?P<ad>\d+):gx\s+(?P<val>[.\d-]+)$", re.I)
  GetGxRe = re.compile(r":conf(?P<ad>\d+):gx\?$", re.I)
  SetA1Re = re.compile(r":conf(?P<ad>\d+):a1\s+(?P<val>[.\d-]+)$", re.I)
  GetA1Re = re.compile(r":conf(?P<ad>\d+):a1\?$", re.I)
  SetA2Re = re.compile(r":conf(?P<ad>\d+):a2\s+(?P<val>[\d-]+)$", re.I)
  GetA2Re = re.compile(r":conf(?P<ad>\d+):a2\?$", re.I)
  SetB1Re = re.compile(r":conf(?P<ad>\d+):b1\s+(?P<val>[.\d-]+)$", re.I)
  GetB1Re = re.compile(r":conf(?P<ad>\d+):b1\?$", re.I)
  SetB2Re = re.compile(r":conf(?P<ad>\d+):b2\s+(?P<val>[\d-]+)$", re.I)
  GetB2Re = re.compile(r":conf(?P<ad>\d+):b2\?$", re.I)

  SetGyRe = re.compile(r":conf(?P<ad>\d+):gy\s+(?P<val>[.\d-]+)$", re.I)
  GetGyRe = re.compile(r":conf(?P<ad>\d+):gy\?$", re.I)
  SetC1Re = re.compile(r":conf(?P<ad>\d+):c1\s+(?P<val>[.\d-]+)$", re.I)
  GetC1Re = re.compile(r":conf(?P<ad>\d+):c1\?$", re.I)
  SetC2Re = re.compile(r":conf(?P<ad>\d+):c2\s+(?P<val>[\d-]+)$", re.I)
  GetC2Re = re.compile(r":conf(?P<ad>\d+):c2\?$", re.I)
  SetD1Re = re.compile(r":conf(?P<ad>\d+):d1\s+(?P<val>[.\d-]+)$", re.I)
  GetD1Re = re.compile(r":conf(?P<ad>\d+):d1\?$", re.I)
  SetD2Re = re.compile(r":conf(?P<ad>\d+):d2\s+(?P<val>[\d-]+)$", re.I)
  GetD2Re = re.compile(r":conf(?P<ad>\d+):d2\?$", re.I)

  def __init__(self, (conn, addr), devices, debugOutputEnabled=False):
    asynchat.async_chat.__init__(self, conn)
    self.devices = devices
    self.debugOutputEnabled = debugOutputEnabled
    self.set_terminator("\n")
    self.outputTerminator = "\n"
    self.terminatorAck = "\006"
    self.buffer = ""

  def collect_incoming_data(self, data):
    self.buffer = self.buffer + data

  def found_terminator(self):
    data = self.buffer
    self.buffer = ""
    self.push(self.terminatorAck)
    if self.debugOutputEnabled:
      sys.stderr.write("< \"%s\"\n" % \
        (data + self.get_terminator()).encode("string_escape"))
      sys.stderr.write("> \"%s\"\n" % \
        self.terminatorAck.encode("string_escape"))
    self.handleClientRequest(data)

  def handleClientRequest(self, request):
    request = request.strip()
    if self.GetXyRe.match(request):
      match = self.GetXyRe.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      self.sendClientResponse("%.5f %.5f" % \
        (self.devices[ad].x, self.devices[ad].y))
    elif self.GetCurrAllRe.match(request):
      match = self.GetCurrAllRe.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      self.sendClientResponse("%d %d %d %d" % (self.devices[ad].curr[0],
        self.devices[ad].curr[1], self.devices[ad].curr[2],
        self.devices[ad].curr[3]))
    elif self.SetAvgCurrRe.match(request):
      match = self.SetAvgCurrRe.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      val = int(match.group("val"))
      if not self.validateAveragingNumValues(val): return
      # nothing more to do - don't support averaging in this simulation
    elif self.SetSingleRe.match(request):
      match = self.SetSingleRe.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      # nothing more to do - don't support averaging in this simulation
    elif self.SetRangeRe.match(request):
      match = self.SetRangeRe.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      val = int(match.group("val"))
      if not self.validateRange(val): return
      self.devices[ad].range = val
    elif self.GetRangeRe.match(request):
      match = self.GetRangeRe.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      self.sendClientResponse("%d" % self.devices[ad].range)
    elif self.SetGxRe.match(request):
      match = self.SetGxRe.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      val = float(match.group("val"))
      self.devices[ad].gx = val
      self.devices[ad].computeXY()
    elif self.GetGxRe.match(request):
      match = self.GetGxRe.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      self.sendClientResponse("%.5f" % self.devices[ad].gx)
    elif self.SetA1Re.match(request):
      match = self.SetA1Re.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      val = float(match.group("val"))
      self.devices[ad].a1 = val
      self.devices[ad].computeXY()
    elif self.GetA1Re.match(request):
      match = self.GetA1Re.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      self.sendClientResponse("%.5f" % self.devices[ad].a1)
    elif self.SetA2Re.match(request):
      match = self.SetA2Re.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      val = int(match.group("val"))
      self.devices[ad].a2 = val
      self.devices[ad].computeXY()
    elif self.GetA2Re.match(request):
      match = self.GetA2Re.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      self.sendClientResponse("%d" % self.devices[ad].a2)
    elif self.SetB1Re.match(request):
      match = self.SetB1Re.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      val = float(match.group("val"))
      self.devices[ad].b1 = val
      self.devices[ad].computeXY()
    elif self.GetB1Re.match(request):
      match = self.GetB1Re.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      self.sendClientResponse("%.5f" % self.devices[ad].b1)
    elif self.SetB2Re.match(request):
      match = self.SetB2Re.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      val = int(match.group("val"))
      self.devices[ad].b2 = val
      self.devices[ad].computeXY()
    elif self.GetB2Re.match(request):
      match = self.GetB2Re.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      self.sendClientResponse("%d" % self.devices[ad].b2)
    elif self.SetGyRe.match(request):
      match = self.SetGyRe.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      val = float(match.group("val"))
      self.devices[ad].gy = val
      self.devices[ad].computeXY()
    elif self.GetGyRe.match(request):
      match = self.GetGyRe.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      self.sendClientResponse("%.5f" % self.devices[ad].gy)
    elif self.SetC1Re.match(request):
      match = self.SetC1Re.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      val = float(match.group("val"))
      self.devices[ad].c1 = val
      self.devices[ad].computeXY()
    elif self.GetC1Re.match(request):
      match = self.GetC1Re.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      self.sendClientResponse("%.5f" % self.devices[ad].c1)
    elif self.SetC2Re.match(request):
      match = self.SetC2Re.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      val = int(match.group("val"))
      self.devices[ad].c2 = val
      self.devices[ad].computeXY()
    elif self.GetC2Re.match(request):
      match = self.GetC2Re.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      self.sendClientResponse("%d" % self.devices[ad].c2)
    elif self.SetD1Re.match(request):
      match = self.SetD1Re.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      val = float(match.group("val"))
      self.devices[ad].d1 = val
      self.devices[ad].computeXY()
    elif self.GetD1Re.match(request):
      match = self.GetD1Re.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      self.sendClientResponse("%.5f" % self.devices[ad].d1)
    elif self.SetD2Re.match(request):
      match = self.SetD2Re.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      val = int(match.group("val"))
      self.devices[ad].d2 = val
      self.devices[ad].computeXY()
    elif self.GetD2Re.match(request):
      match = self.GetD2Re.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      self.sendClientResponse("%d" % self.devices[ad].d2)
    elif self.SetIcRe.match(request):
      match = self.SetIcRe.match(request)
      ad = int(match.group("ad"))
      if not self.validateAddress(ad): return
      ch = int(match.group("ch"))
      if not self.validateChannel(ch): return
      val = int(match.group("val"))
      self.devices[ad].curr[ch - 1] = val
      self.devices[ad].computeXY()
    else:
      sys.stderr.write("error: unknown command \"%s\"\n" % \
        request.encode("string_escape"))

  def sendClientResponse(self, response):
    data = response + self.outputTerminator
    self.push(data)
    if self.debugOutputEnabled:
      sys.stderr.write("> \"%s\"\n" % data.encode("string_escape"))

  def validateAddress(self, ad):
      if ad >= 0 and ad <= 15: return True
      sys.stderr.write("error: invalid device address \"%d\"\n" % ad)
      return False

  def validateAveragingNumValues(self, num):
      if num >= 2 and num <= 99: return True
      sys.stderr.write("error: invalid averaging num values \"%d\"\n" % num)
      return False

  def validateChannel(self, ch):
      if ch >= 1 and ch <= 4: return True
      sys.stderr.write("error: invalid input current channel \"%d\"\n" % ch)
      return False

  def validateRange(self, rng):
      if rng >= 1 and rng <= 6: return True
      sys.stderr.write("error: invalid input current range \"%d\"\n" % rng)
      return False


class IcPlusQuad:
  def __init__(self):
    self.x = 0.0
    self.y = 0.0
    self.curr = [0, 0, 0, 0]

    self.range = 1

    self.gx = 1.0
    self.a1 = 1.0
    self.a2 = 0
    self.b1 = 1.0
    self.b2 = 0

    self.gy = 1.0
    self.c1 = 1.0
    self.c2 = 0
    self.d1 = 1.0
    self.d2 = 0

  def computeXY(self):
    a = self.a1 * float(self.curr[3] - self.a2)
    b = self.b1 * float(self.curr[1] - self.b2)
    c = self.c1 * float(self.curr[0] - self.c2)
    d = self.d1 * float(self.curr[2] - self.d2)

    if a + b != 0.0: self.x = self.gx * (a - b) / (a + b)
    if c + d != 0.0: self.y = self.gy * (c - d) / (c + d)


if __name__ == '__main__':
  try:
    initializeSignalHandlers()
    main(sys.argv)
  except Exception, e:
    if isinstance(e, SystemExit):
      raise e
    else:
      print >> sys.stderr
      print >> sys.stderr, "Error: %s" % e
      sys.exit(1)
